
from lib import *

class ResBlock(T.nn.Module):
    def __init__(self, c_in, c_out, sampler):
        super().__init__()
        
        self.conv = T.nn.Sequential(*[T.nn.Conv2d(c_in, c_out, 3, padding=1), T.nn.ReLU(), 
                                      sampler, 
                                      T.nn.Conv2d(c_out, c_out, 3, padding=1)])
        self.skip = T.nn.Sequential(*[T.nn.Conv2d(c_in, c_out, 1, padding=0), 
                                      sampler])
    
    def forward(self, inp):
        conv, skip = self.conv(inp), self.skip(inp)
        out = conv+skip
        
        return out
    
class Encoder(T.nn.Module):
    def __init__(self):
        super().__init__()
        
        self.rgb = T.nn.Conv2d(3, 32, 1, padding=0, stride=2)
        self.res = T.nn.Sequential(*[T.nn.ReLU(), ResBlock(32, 64, T.nn.AvgPool2d(2)), T.nn.BatchNorm2d(64), 
                                     T.nn.ReLU(), ResBlock(64, 128, T.nn.AvgPool2d(2)), T.nn.BatchNorm2d(128), 
                                     T.nn.ReLU(), ResBlock(128, 256, T.nn.AvgPool2d(2))])
    def forward(self, inp):
        out = self.rgb(inp)
        out = self.res(out)
        
        return out
    
class SelfAttention(T.nn.Module):
    def __init__(self, c):
        super().__init__()
        
        self.f, self.g, self.h = T.nn.Conv2d(c, c//8, 1, padding=0), T.nn.Conv2d(c, c//8, 1, padding=0), T.nn.Conv2d(c, c//2, 1, padding=0)
        self.c = T.nn.Conv2d(c, c//2, 1, padding=0)
        
    def forward(self, inp):
        B, C, H, W = inp.shape
        N = H*W
        
        f, g = self.f(inp).view(B, C//8, N), self.g(inp).view(B, C//8, N)
        a = T.nn.functional.softmax(T.bmm(f.permute(0, 2, 1), g), dim=1)
        
        h = self.h(inp).view(B, C//2, N)
        o = T.bmm(h, a).view(B, C//2, H, W)
        
        out = o+self.c(inp)
        
        return out
    
class Decoder(T.nn.Module):
    def __init__(self):
        super().__init__()
        
        self.res = T.nn.Sequential(*[T.nn.ReLU(), ResBlock(256, 128, T.nn.Upsample(scale_factor=2)), T.nn.BatchNorm2d(128), 
                                     T.nn.ReLU(), ResBlock(128, 64, T.nn.Upsample(scale_factor=2)), T.nn.BatchNorm2d(64), 
                                     T.nn.ReLU(), ResBlock(64, 32, T.nn.Upsample(scale_factor=2))])
        self.rgb = T.nn.Sequential(*[T.nn.ReLU(), T.nn.Upsample(scale_factor=2), T.nn.Conv2d(32, 3, 1, padding=0), 
                                     T.nn.Sigmoid()])
        
    def forward(self, inp):
        out = self.res(inp)
        out = self.rgb(out)
        
        return out

class Model(T.nn.Module):
    def __init__(self):
        super().__init__()
        
        self.enc, self.dec = Encoder(), Decoder()
        self.sa = SelfAttention(256+256)
        self.i2f = T.nn.Sequential(*[T.nn.Linear(1024, 512), T.nn.ReLU(), T.nn.Linear(512, 256)])
    
    def forward_enc(self, inp):
        hid = self.enc(inp)
        feat = T.nn.functional.adaptive_avg_pool2d(hid, (1, 1)).squeeze()
        
        return hid, feat
    
    def forward_dec(self, feat_sty, hid_con):
        feat_sty = feat_sty.unsqueeze(-1).unsqueeze(-1).expand(-1, -1, hid_con.shape[2], hid_con.shape[3])
        hid = T.cat([feat_sty, hid_con], dim=1)
        hid = self.sa(hid)
        out = self.dec(hid)
        
        return out
    
if __name__=='__main__':
    model = Model().cuda()
    
    _, feat_sty = model.forward_enc(T.rand(50, 3, 128, 128).cuda())
    hid_con, _ = model.forward_enc(T.rand(50, 3, 192, 256).cuda())
    out = model.forward_dec(feat_sty, hid_con)
    print(out.shape)
    
